!=========================================================================
! Copyright (C) GemTalk Systems 1986-2016.  All Rights Reserved.
!
! $Id: gsfile.gs,v 1.7 2008-01-09 19:02:33 stever Exp $
!
! Superclass Hierarchy:
!   GsFile, Object.
!
!=========================================================================

expectvalue %String
run
^ Object _newKernelSubclass: 'GsFile'
        instVarNames: #(  
          #isClient "a Boolean, true means file is on client"
          #pathName  "a String, directory plus file name"
          #mode      "a String")

        classVars: #()
        classInstVars: #()
        poolDictionaries: #[]
        inDictionary: Globals
        constraints: #[#[ #isClient, Boolean ],
                       #[ #pathName, String ], 
		       #[ #mode, String ]  ]
        instancesInvariant: false
        isModifiable: false
        reservedOop: 1797
%

! Remove existing behavior from GsFile
removeallmethods GsFile
removeallclassmethods GsFile

category: 'For Documentation Installation only'
classmethod: GsFile
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'GsFile provides the means for creating and accessing non-GemStone files.  Such
 files reside in the file system on either the machine that is running the
 current session''s Gem process (the server machine) or the machine that is
 running the client application (the client machine).  The files may be of any
 type, textual or binary, though separate protocol is provided for reading and
 writing these types of data.

                              Warning:
    Do not retain an instance of GsFile from one session to another.
    Instances of GsFile are intended to exist only within a given GemStone
    session.  GsFiles that are used across sessions always generate an error.'.
doc documentClassWith: txt.

txt := (GsDocText new) details:
'A Boolean whose value is true if the file is on the client machine and false
 if it is on the server machine.' .
doc documentInstVar: #isClient with: txt.

txt := (GsDocText new) details:
'A String that gives an absolute path name to the file.' .
doc documentInstVar: #pathName with: txt.

txt := (GsDocText new) details:
'A String that gives the file open mode for the file, as defined for the
 C-language open function.' .
doc documentInstVar: #mode with: txt.

self description: doc.
%

! ------------------- Class methods for GsFile
category: 'Drastic Measures'
classmethod: GsFile
closeAll

"Closes all open files on the client machine except stdin/stdout/stderr.
 Returns the receiver if successful, nil if not."

"make sure the sources file has been explicitly closed"

^ self classUserAction: #GsfCloseAll onClient: true
%

category: 'Drastic Measures'
classmethod: GsFile
closeAllOnServer

"Closes all open files on the server machine except stdin/stdout/stderr.
 Returns the receiver if successful, nil if not."

^ self classUserAction: #GsfCloseAll onClient: false
%

category: 'Garbage Collection'
classmethod: GsFile
finalizeAll

"Closes all open files on the client machine that were opened by a
 instance of GsFile but that are no longer being used by an instance
 of GsFile. The need for finalization is caused by instances being
 garbage collected before they are sent close.
 Returns nil if an error occurs.
 Returns the total number of GsFile instances that were not finalized
 because they were still in use."

System _generationScavenge.
^ self classUserAction: #GsfFinalizeAll onClient: true
%

category: 'Garbage Collection'
classmethod: GsFile
finalizeAllOnServer

"Closes all open files on the server machine that were opened by a
 instance of GsFile but that are no longer being used by an instance
 of GsFile. The need for finalization is caused by instances being
 garbage collected before they are sent close.
 Returns nil if an error occurs.
 Returns the total number of GsFile instances that were not finalized
 because they were still in use."

System _generationScavenge.
^ self classUserAction: #GsfFinalizeAll onClient: false
%

category: 'Directory Operations'
classmethod: GsFile
contentsAndTypesOfDirectory: dirSpec onClient: bool

"Returns an Array of objects describing the contents of the given directory.
 The argument bool indicates the location of the directory as on the client or
 the server.  Successive pairs of elements of the Array are each the name of an
 entry, and a Boolean - true if the entry is a file, and false if not.

 Sample:  #['file.c', true, 'subdir', false, ...]

 Returns anArray if successful, nil if not."

| aLine dirContents result |

dirContents := self _directory: dirSpec onClient: bool .
dirContents == nil ifTrue:[ ^ nil ].

result := Array new .
[ aLine := dirContents nextLine .  aLine == nil ] whileFalse: [ 
  aLine isEmpty
    ifFalse: [
      aLine last == Character lf ifTrue:[ aLine size: (aLine size - 1) ].
      result add: aLine .
      result add: (self _fileKind: aLine onClient: bool) == 0  .
    ]
  ] .
dirContents close .
self _removeFile: dirContents pathName onClient: bool .
^ result
%

category: 'Private'
classmethod: GsFile
_fileKind: aPathName onClient: clientBool

"Returns a SmallInteger representing the kind of the file, or returns nil if an
 error occurs, in which case the class error buffer contains the error.

 The file kinds are enumerated as follows:

    0. file
    1. directory
    2. character device
    3. block device
    4. symbolic link
    5. other
    6. error"

^ self classUserAction: #GsfFileKind onClient: clientBool with: aPathName
%

category: 'Directory Operations'
classmethod: GsFile
contentsOfDirectory: dirSpec onClient: bool

"Returns an Array of Strings describing the contents of the given directory.
 The argument bool indicates the location of the directory as on the client or
 the server.  Each element of the Array is the name of an entry.

 Returns anArray if successful, nil if not."

| aLine dirContents result |

dirContents := self _directory: dirSpec onClient: bool .
dirContents == nil ifTrue:[ ^ nil ].

result := Array new .
[ aLine := dirContents nextLine .  aLine == nil ] whileFalse: [ 
  aLine isEmpty
    ifFalse: [
      aLine last == Character lf ifTrue:[ aLine size: (aLine size - 1) ].
      result add: aLine 
    ]
  ] .
dirContents close .
self _removeFile: dirContents pathName onClient: bool .
^ result
%

category: 'Private'
classmethod: GsFile
_directory: dirSpec onClient: clientBool

"Returns an opened temporary text file containing the contents of the specified
 directory, one line per directory entry."

| dirFileName |

dirFileName := self classUserAction: #GsfDirectory onClient: clientBool
                                                   with: dirSpec .
dirFileName == nil ifTrue: [ ^ nil ].
^ self open: dirFileName mode: 'r' onClient: clientBool.
%

category: 'File Operations'
classmethod: GsFile
exists: aPathName

"Returns true if the given path points to a file on the client,
 false if not, and nil if an error occurs trying to find out."

^ self _exists: aPathName onClient: true
%

category: 'File Operations'
classmethod: GsFile
existsOnServer: aPathName

"Returns true if the given path points to a file, on the server,
 false if not, and nil if an error occurs trying to find out."

^ self _exists: aPathName onClient: false
%

category: 'Private'
classmethod: GsFile
_exists: aPathName onClient: clientBool

"Returns true if the given path points to a file, false if not, and nil if an
 error other than file-does-not-exist occurs trying to find out."

| result |

result := self classUserAction: #GsfSize onClient: clientBool with: aPathName .
result == -1 ifTrue:[ ^ false "file does not exist" ].
result == nil ifTrue:[ ^ nil "some other error" ] .
^ true
%

category: 'Instance Creation'
classmethod: GsFile
open: aPathName mode: openMode 

"Creates an instance of the receiver and opens a file on the client machine.

 The openMode argument must be a String that that is equal to one of the
 following: 'r', 'w', 'a', 'r+', 'w+', 'a+', 'rb', 'wb', 'ab', 'r+b', 'w+b',
 'a+b', 'rb+', 'wb+', 'ab+'.  The mode has the same meaning as it does for the
 C library function, fopen().

 Returns a GsFile if successful, nil if an error occurs."

^ self open: aPathName mode: openMode onClient: true
%

category: 'Instance Creation'
classmethod: GsFile
newWithFilePtr: filePtr pathname: aPathName mode: openMode onClient: clientBool 

"Creates an instance of the receiver using an already open FILE*: filePtr.
 aPathName, openMode, and clientBool describe the already open filePtr.
 The filePtr most likely comes from a C user action. This method assumes
 it was converted to an object using GciLongToOop().
 
 Returns a GsFile if successful, nil if an error occurs."

| inst |

inst := self new.
inst _open: aPathName mode: openMode onClient: clientBool.
^ inst userAction: #GsfAssociate onClient: clientBool with: filePtr
%

category: 'Instance Creation'
classmethod: GsFile
open: aPathName mode: openMode  onClient: clientBool 

"Creates an instance of the receiver and opens a file on the client machine
 (if clientBool is true) or the server machine (if clientBool is false).

 The openMode argument must be a String that that is equal to one of the
 following: 'r', 'w', 'a', 'r+', 'w+', 'a+', 'rb', 'wb', 'ab', 'r+b', 'w+b',
 'a+b', 'rb+', 'wb+', 'ab+'.  The mode has the same meaning as it does for the
 C library function, fopen().

 Returns a GsFile if successful, nil if an error occurs."

  | inst |

(#('r'  'w'  'a'  'r+'  'w+'  'a+'
   'rb' 'wb' 'ab' 'r+b' 'w+b' 'a+b'
                  'rb+' 'wb+' 'ab+') includesValue: openMode) ifFalse: [
  ^nil
].
inst := self new.
inst _open: aPathName mode: openMode onClient: clientBool.
^ inst open
%

category: 'Instance Creation'
classmethod: GsFile
openOnServer: aPathName mode: openMode

"Creates an instance of the receiver and opens a file on the server machine.

 The openMode argument must be a String that that is equal to one of the
 following: 'r', 'w', 'a', 'r+', 'w+', 'a+', 'rb', 'wb', 'ab', 'r+b', 'w+b',
 'a+b', 'rb+', 'wb+', 'ab+'.  The mode has the same meaning as it does for the
 C library function, fopen().

 Returns a GsFile if successful, nil if an error occurs."

^ self open: aPathName mode: openMode onClient: false
%

category: 'Instance Creation'
classmethod: GsFile
openAppend: aPathName

"Opens a text file on the client machine for writing.
 If the file does not exist it is created.
 The file position indicator is positioned at the end of the file
 before each write operation.
 Read operations are not allowed.
 Returns aGsFile if successful, nil if an error occurs."

^self open: aPathName mode: 'a' onClient: true
%

category: 'Instance Creation'
classmethod: GsFile
openRead: aPathName

"Opens an existing text file on the client machine for reading.
 Write operations are not allowed.
 Returns aGsFile if successful, nil if an error occurs."

^self open: aPathName mode: 'r' onClient: true
%

category: 'Instance Creation'
classmethod: GsFile
openUpdate: aPathName

"Opens an existing text file on the client machine for reading and writing.
 Returns aGsFile if successful, nil if an error occurs."

^self open: aPathName mode: 'r+' onClient: true
%

category: 'Instance Creation'
classmethod: GsFile
openWrite: aPathName

"Opens a text file on the client machine for writing.
 If the file exists it is truncated to zero length.
 If the file does not exist it is created.
 Read operations are not allowed.
 Returns aGsFile if successful, nil if an error occurs."

^self open: aPathName mode: 'w' onClient: true
%

category: 'Instance Creation'
classmethod: GsFile
openAppendOnServer: aPathName

"Opens a text file on the server machine for writing.
 If the file does not exist it is created.
 The file position indicator is positioned at the end of the file
 before each write operation.
 Read operations are not allowed.
 Returns aGsFile if successful, nil if an error occurs."

^self open: aPathName mode: 'a' onClient: false
%

category: 'Instance Creation'
classmethod: GsFile
openReadOnServer: aPathName

"Opens an existing text file on the server machine for reading.
 Write operations are not allowed.
 Returns aGsFile if successful, nil if an error occurs."

^self open: aPathName mode: 'r' onClient: false
%

category: 'Instance Creation'
classmethod: GsFile
openUpdateOnServer: aPathName

"Opens an existing text file on the server machine for reading and writing.
 Returns aGsFile if successful, nil if an error occurs."

^self open: aPathName mode: 'r+' onClient: false
%

category: 'Instance Creation'
classmethod: GsFile
openWriteOnServer: aPathName

"Opens a text file on the server machine for writing.
 If the file exists it is truncated to zero length.
 If the file does not exist it is created.
 Read operations are not allowed.
 Returns aGsFile if successful, nil if an error occurs."

^self open: aPathName mode: 'w' onClient: false
%

category: 'File Operations'
classmethod: GsFile
removeClientFile: aPathName

"Removes the named file from the client machine's file system.
 Returns the receiver if the file was deleted, nil if an error occurs."

^ self _removeFile: aPathName onClient: true .
%

category: 'File Operations'
classmethod: GsFile
removeServerFile: aPathName

"Removes the named file from the server machine.  Returns the receiver if the
 file was deleted, nil if an error occurs."

^ self _removeFile: aPathName onClient: false .
%

category: 'Private'
classmethod: GsFile
_removeFile: aPathName onClient: clientBool

"Returns receiver if removal succeeds.  Returns nil if an error occurs,
 in which case the class error buffer contains the error."

^ self classUserAction: #GsfUnlink onClient: clientBool with: aPathName
%

category: 'Directory Operations'
classmethod: GsFile
removeClientDirectory: aPathName

"Removes the named directory from the client machine's file system.
 Returns the receiver if the directory was deleted, nil if an error occurs.
 Note that this method will fail if the directory is not empty."

^ self _removeDirectory: aPathName onClient: true .
%

category: 'Directory Operations'
classmethod: GsFile
removeServerDirectory: aPathName

"Removes the named directory from the server machine.
 Returns the receiver if the directory was deleted, nil if an error occurs.
 Note that this method will fail if the directory is not empty."

^ self _removeDirectory: aPathName onClient: false .
%

category: 'Private'
classmethod: GsFile
_removeDirectory: aPathName onClient: clientBool

"Returns receiver if removal succeeds.  Returns nil if an error occurs,
 in which case the class error buffer contains the error."

^ self classUserAction: #GsfRmdir onClient: clientBool with: aPathName
%

category: 'Directory Operations'
classmethod: GsFile
isClientDirectory: aPathName
"Returns true if 'aPathName' names an existing directory on the client
 machine.
 Returns false if it is not a directory.
 Returns nil if an error occurs."

^ self _isDirectory: aPathName onClient: true .
%

category: 'Directory Operations'
classmethod: GsFile
isServerDirectory: aPathName
"Returns true if 'aPathName' names an existing directory on the server
 machine.
 Returns false if it is not a directory.
 Returns nil if an error occurs."

^ self _isDirectory: aPathName onClient: false .
%

category: 'Private'
classmethod: GsFile
_isDirectory: aPathName onClient: clientBool

"Returns true if 'aPathName' names an existing directory.
 Returns false if it is not a directory.
 Returns nil if an error occurs."

^ self classUserAction: #GsfIsDir onClient: clientBool with: aPathName
%

category: 'Directory Operations'
classmethod: GsFile
createClientDirectory: aPathName

"Creates the named directory on the client machine's file system.
 Returns the receiver if the directory was created, nil if an error occurs."

^ self _createDirectory: aPathName onClient: true .
%

category: 'Directory Operations'
classmethod: GsFile
createServerDirectory: aPathName

"Creates the named directory on the server machine.
 Returns the receiver if the directory was created, nil if an error occurs."

^ self _createDirectory: aPathName onClient: false .
%

category: 'Private'
classmethod: GsFile
_createDirectory: aPathName onClient: clientBool

"Returns receiver if creation succeeds.  Returns nil if an error occurs,
 in which case the class error buffer contains the error."

^ self classUserAction: #GsfMkdir onClient: clientBool with: aPathName
%

category: 'File Operations'
classmethod: GsFile
lastModificationOfClientFile: aPathName

"Gets the date and time the named file on the client machine was last
 modified. Returns a DateTime if successful, nil if an error occurs."

^ self _fileModTime: aPathName onClient: true .
%

category: 'File Operations'
classmethod: GsFile
lastModificationOfServerFile: aPathName

"Gets the date and time the named file on the server machine was last 
 modified. Returns a DateTime if successful, nil if an error occurs."

^ self _fileModTime: aPathName onClient: false .
%

category: 'Accessing'
method: GsFile
lastModified
"Returns a DateTime that represents the last time the receiver's file
 was modified. Returns nil if an error occurs."

^ GsFile _fileModTime: (self pathName) onClient: (self isClient).
%

category: 'Private'
classmethod: GsFile
_fileModTime: aPathName onClient: clientBool

"Returns DateTime of files last modification time if successful.
 Returns nil if an error occurs,
 in which case the class error buffer contains the error."

^ self classUserAction: #GsfModTime onClient: clientBool with: aPathName
%

category: 'File Operations'
classmethod: GsFile
sizeOf: aPathName

"Returns the size in bytes of the given client file, or nil if an error occurs."

^ self _sizeOf: aPathName onClient: true
%

category: 'File Operations'
classmethod: GsFile
sizeOfOnServer: aPathName

"Returns the size in bytes of the given server file, or nil if an error occurs."

^ self _sizeOf: aPathName onClient: false
%

category: 'Private'
classmethod: GsFile
_sizeOf: aPathName onClient: clientBool

"Returns size of the file in bytes.  Returns nil if an error occurs,
 in which case the class error buffer contains the error."

| result |

result := self classUserAction: #GsfSize onClient: clientBool with: aPathName .
result == nil ifTrue:[ ^ nil ] .
result == -1 ifTrue:   "file does not exist"
  ["to get error into the buffer"
   self _fileModTime: aPathName onClient: clientBool.
   ^nil ] .
^ result
%

category: 'Standard Files'
classmethod: GsFile
stdin

"Returns an instance of the receiver that is set up to read the standard input
 of the client process, or nil if an error occurs."

^self _getStdFile: 0 onClient: true
%

category: 'Standard Files'
classmethod: GsFile
stdout

"Returns an instance of the receiver that is set up to write to the standard
 output of the client process, or nil if an error occurs."

^self _getStdFile: 1 onClient: true
%

category: 'Standard Files'
classmethod: GsFile
stderr

"Returns an instance of the receiver that is set up to write to the standard
 error output of the client process, or nil if an error occurs."

^self _getStdFile: 2 onClient: true
%

category: 'Standard Files'
classmethod: GsFile
stdoutServer

"Returns an instance of the receiver that is set up to write to the standard
 output of the server process, or nil if an error occurs."

^self _getStdFile: 1 onClient: false
%


category: 'Private'
classmethod: GsFile
_getStdFile: stdId onClient: clientBool

"Returns an instance of the receiver that is set up to operate on a standard
 file or nil if an error occurs.  stdId can be 0 (stdin),
 1 (stdout), or 2 (stderr)."

|result stdIdCache virtualId |

stdIdCache := System _sessionStateAt: 18.
(stdIdCache == nil) ifTrue: [
  stdIdCache := Array new: 6.
  System _sessionStateAt: 18 put: stdIdCache.
].
clientBool ifTrue:[   virtualId := stdId + 1 ]
          ifFalse:[  virtualId := stdId + 4 ].
result := stdIdCache at: virtualId .
result == nil ifTrue: [ | ok |
  "It has not yet been cached so create it"
  result := self new.
  result _newStdFile: stdId isClient: clientBool .
  "tell the client about it"
  ok := result userAction: #GsfCreateStdFile onClient: clientBool with: stdId.
  ok ifTrue:[ stdIdCache at: virtualId put: result.]
    ifFalse:[ result := nil. ] .
].
^ result
%

category: 'Error Reporting'
classmethod: GsFile
lastErrorString

"Returns the currently posted error string, for class operations on the
 client, or nil if no error has occurred.  Clears the error string."

^ self classUserAction: #GsfClassError onClient: true
%

category: 'Error Reporting'
classmethod: GsFile
serverErrorString

"Returns the currently posted error string, for class operations on the
 server, or nil if no error has occurred.  Clears the error string."

^ self classUserAction: #GsfClassError onClient: false
%

category: 'Private'
classmethod: GsFile
_clientVersionAt: verParamId

"Returns client process version information.  verInfoId should be one of the
 values in the global dictionary VersionParameterDict."

^ self classUserAction: #GsfVersionParam onClient: true with: verParamId
%

category: 'Private'
classmethod: GsFile
_expandEnvVariable: varName isClient: clientBool

"Expands the environment variable named varName in either the GemBuilder for C
 or Gem process, returning a String. varName should be a kind of String.

 Returns nil if any of the following are true
    varName is not a byte format object.
    there is no environment variable defined with name  varName,
    the value of the environment variable is more than approximately 8000 bytes,
    the size of varName exceeds approximately 8000 bytes."

^ self classUserAction: #GsfExpandEnvVar onClient: clientBool with: varName
%

category: 'Private'
classmethod: GsFile
classUserAction: actionName onClient: onClient

"Executes a GsFile user action, passing it the following arguments:
    self
    any arguments to the primitive (see other uses of primitive 396)
 Maximum of 6 arguments (6 with: keywords) for this primitive."

<primitive: 396>

^ self _primitiveFailed: #classUserAction:onClient:
%

category: 'Private'
classmethod: GsFile
classUserAction: actionName onClient: onClient with: arg1

"See GsFile | classUserAction:onClient: for documentation."

<primitive: 396>

^ self _primitiveFailed: #classUserAction:onClient:with:
%

category: 'Private'
classmethod: GsFile
_log: string
  "Convenience method to print to stdout just like log: does."

  self stdout log: string
%

! ------------------- Instance methods for GsFile

category: 'Private'
method: GsFile
_open: aPath mode: aMode  onClient: clientBool 

"Initialize and open the receiver as specified by the arguments."

isClient := clientBool .  
pathName := aPath .
mode := aMode .
%

category: 'File Operations'
method: GsFile
fileSize

"Returns the size in bytes of the receiver, or nil if an error occurs.

 Note that the value returned is independent of the open mode used to
 create the receiver."

^ self class _sizeOf: pathName onClient: isClient
%

category: 'Writing'
method: GsFile
+ collection

"Writes the contents of the given collection to the receiver's file at
 the current position. The argument must be a kind of CharacterCollection
 with byte format.  Returns a count of bytes added, or nil if an error occurs."

^  self nextPutAll: collection
%

category: 'Writing'
method: GsFile
add: char

"Writes the given Character to the receiver's file at the current position. 
 Returns true, or nil if an error occurs."

^ self nextPut: char
%

category: 'Writing'
method: GsFile
addAll: collection

"Writes the contents of the given collection to the receiver's file at
 the current position. The argument must be a kind of CharacterCollection
 with byte format.  Returns a count of bytes added, or nil if an error occurs."

^ self nextPutAll: collection
%

category: 'Positioning'
method: GsFile
atEnd

"Returns true if the receiver is currently positioned at the end of its
 file, false if not, or nil if an error occurs."

^ self userAction: #GsfAtend onClient: isClient
%

category: 'File Operations'
method: GsFile
close

"Closes the receiver's file.  Returns the receiver, or nil if an error occurs."

self isOpen ifFalse: [^self].
^ self userAction: #GsfClose onClient: isClient
%

category: 'Reading'
method: GsFile
contents

"Returns a String containing the contents of the receiver from the current
 position to the end of the file.  Returns nil if an error occurs."

| mySize myPosition |

mySize := self fileSize .
mySize == nil ifTrue:[ ^ nil ] .
myPosition := self position .
myPosition == nil ifTrue:[ ^ nil ] .
^ self next: ( mySize - myPosition ) 
%

category: 'Private'
method: GsFile
_isBinary

"Returns true if receiver is a binary file. Otherwise returns false."
(mode == nil) ifTrue: [^false].
^mode includesValue: $b
%

category: 'Writing'
method: GsFile
cr

"If the receiver is a text file then writes an end-of-line sequence to it.
 If the receiver is a binary file then writes a carriage return to it.
 Returns a count of bytes added, or nil if an error occurs."

(self _isBinary) ifTrue: [
  ^ self userAction: #GsfPutC onClient: isClient with: 13 "ASCII carriage return"
] ifFalse: [
  "ANSI-C text files will write eoln when they see a single lf character."
  ^ self lf
].
%

category: 'Error Reporting'
method: GsFile
lastErrorString

"Returns the currently posted error string, or nil if no error has occurred.
 Clears the error string and error code."

^ self userAction: #GsfErrorString onClient: isClient
%

category: 'Error Reporting'
method: GsFile
lastErrorCode

"Returns the currently posted error code, or zero if no error has occurred.
 Does not clear the error code or error string."

^ self userAction: #GsfErrorCode onClient: isClient
%

category: 'Writing'
method: GsFile
ff

"Writes a form-feed (page break) to the receiver's file.
 Returns true, or nil if an error occurs."

^ self nextPut:  Character newPage
%

category: 'File Operations'
method: GsFile
flush

"Flushes all written bytes to the file.  Returns the receiver, or nil if an
 error occurs."

^ self userAction: #GsfFlush onClient: isClient
%

category: 'File Operations'
method: GsFile
head: lineCount

"Returns a String containing the first lineCount lines from the
 receiver's file, or nil if an error occurs."

| result line |

self position: 0.
result := String new.
lineCount timesRepeat: [
  line := self nextLine.
  line ~~ nil ifTrue: [
    result addAll: line.
  ]
  ifFalse: [
    ^result
  ]
].
^result
%

category: 'Testing'
method: GsFile
isExternal

"Is the source for this stream is external to GemStone Smalltalk."

^true
%

category: 'Testing'
method: GsFile
isClient

"Returns true if the receiver's file is a client file, or nil if an error
 occurs."

^ isClient.
%

category: 'Testing'
method: GsFile
isOpen

"Returns true if the receiver's file is open, or nil if an error occurs."

^ (self id ~~ 0)
%

category: 'Writing'
method: GsFile
lf

"If the receiver is a text file then writes an end-of-line sequence to it.
 If the receiver is a binary file then write a line-feed to it.
 Returns a count of bytes added, or nil if an error occurs."

^ self userAction: #GsfPutC onClient: isClient with: 10
                                                  "ASCII line-feed Character"
%

category: 'Writing'
method: GsFile
log: string

"Writes the contents of the given collection to the receiver's file at the
 current position.  Appends a newline to the file if the string does not end
 with one.  The argument must be a kind of CharacterCollection with byte format.

 Returns the receiver if successful; returns nil otherwise."

| lf result |

result := self nextPutAll: string.
(result ~~ nil _and:[ result > 0]) ifTrue:[
  lf := Character lf.
  string last == lf ifFalse: [ result := self nextPut: lf ]
  ].
result == nil ifTrue: [ ^nil ].
self flush.
^ self
%

category: 'Accessing'
method: GsFile
mode

"Returns the access mode of the receiver's file."

^mode
%

category: 'Accessing'
method: GsFile
name

"Returns the receiver's file path name."

^ pathName
%

category: 'Reading'
method: GsFile
next

"Returns the next Character from the receiver's file, or nil if an error
 occurs."

^ self userAction: #GsfGetC onClient: isClient
%

category: 'Reading'
method: GsFile
next: numberOfBytes

"Returns a String containing the next numberOfBytes Characters from the
 receiver's file, or nil if an error occurs."

| result |
result := String new.
(self next: numberOfBytes ofSize: 1 into: result ) == nil ifTrue:[ ^ nil ].
^ result
%

category: 'Reading'
method: GsFile
next: amount byteStringsInto: byteObj

"Reads bytes written by printBytes: into the given byte object.
 Returns count of bytes read, or nil if an error occurs."

^ self userAction: #GsfLoadByteStrings onClient: isClient with: byteObj with: amount
%

category: 'Reading'
method: GsFile
next: numberOfBytes into: aCharacterCollection

"Reads the next numberOfBytes into the given collection object. The
 object's size is truncated to the amount of data actually read.
 Returns a count of bytes read, or nil if an error occurs."

^ self next: numberOfBytes ofSize: 1 into: aCharacterCollection 
%

category: 'Reading'
method: GsFile
_next: numberOfBytes basicInto: aCharacterCollection

""

"GsFile doesn't distinguish Strings from EUCStrings, so no reimplementation
 needed."

^ self next: numberOfBytes into: aCharacterCollection
%

category: 'Reading'
method: GsFile
next: numberOfItems ofSize: bytesPerItem into: byteObj

"Reads bytes for the next numberOfItems of the given bytesPerItem into 
 the given collection object. The object's size is truncated to the 
 amount of data actually read.  bytesPerItem must between 1 and 4096 inclusive.

 Returns a count of bytes read, or nil if an error occurs."

^ self userAction: #GsfRead onClient: isClient 
       with: byteObj with: bytesPerItem with: numberOfItems
%

category: 'Reading'
method: GsFile
nextByte

"Returns the next byte (integer) from the receiver's file, or nil if an
error occurs."

^ self next asciiValue 
%

category: 'Reading'
method: GsFile
nextLine

"Returns a String containing the next line from the receiver's file. The String
 will be terminated with a newline, unless the end of file is reached and there
 is no line terminator.  If the receiver is positioned at the end of the file,
 returns an empty String.  Returns nil if an error occurs."

"There is no limit on line size."

| result |
result := String new .
(self nextLineInto: result startingAt: 1) == nil ifTrue:[ ^ nil ] .
^ result .
%

category: 'Reading'
method: GsFile
nextLineInto: str startingAt: pos

"Reads the next line from the receiver's file into the given collection object,
 starting at the given position in the collection. The collection will be
 terminated with a newline, unless the end of file is reached and there is no
 line terminator.  If the receiver is positioned at the end of the file, nothing
 is written.  Returns a count of bytes read, or nil if an error occurs."

"There is no limit on line size."

^ self userAction: #GsfGetLine onClient: isClient with: str with: pos.
%

category: 'Writing'
method: GsFile
nextPut: aByte

"Writes the given byte to the receiver's file at the current position.
 aByte must be a Character or a SmallInteger in the range 0..255.

 If aByte is a SmallInteger, it will be interpreted logically as

 Character withValue: aByte 
   
 Returns true, or nil if an error occurs."

^ self userAction: #GsfPutC onClient: isClient with: aByte
%

category: 'Writing'
method: GsFile
nextPutAll: collection

"Writes the contents of the given collection to the receiver's file at the
 current position. The argument must be a kind of CharacterCollection with byte
 format.  Returns a count of bytes added, or nil if an error occurs."

^ self write: collection itemCount: (collection _basicSize) ofSize: 1 
%

category: 'Writing'
method: GsFile
nextPutAllBytes: collection

"Writes the byte contents of the given collection to the receiver's file at the
 current position.  The argument must be a kind of CharacterCollection with byte
 format.  Returns a count of bytes added, or nil if an error occurs."

^ self nextPutAll: collection
%

category: 'File Operations'
method: GsFile
open

"If the receiver is not open, open it using the existing mode.
 Returns the receiver, or nil if an error occurs."

(pathName == nil _or: [mode == nil]) ifTrue: [ ^nil ].

(self isOpen) ifTrue: [ ^self ].

^ self userAction: #GsfOpen onClient: isClient with: pathName with: mode
%

category: 'File Operations'
method: GsFile
open: aPathName mode: openMode

"Opens the receiver's file with the given mode.  If the file is already open,
 it is closed and reopened.

 The openMode argument must be a String that that is equal to one of the
 following: 'r', 'w', 'a', 'r+', 'w+', 'a+', 'rb', 'wb', 'ab', 'r+b', 'w+b',
 'a+b', 'rb+', 'wb+', 'ab+'.  The mode has the same meaning as it does for the
 C library function, fopen().

 Returns the receiver if successful, or nil if an error occurs."

self isOpen ifTrue:[ self close ].
mode := openMode.
pathName := aPathName.
^self open
%

category: 'Accessing'
method: GsFile
pathName

"Returns the receiver's file path name."

^pathName
%

category: 'Reading'
method: GsFile
peek

"Returns the next byte in the receiver's file, without advancing the current
 pointer.  Returns nil if an error occurs."

| pos result |

pos := self position.
result := self next.
self position: pos.
^result
%

category: 'Reading'
method: GsFile
peek2

"Returns the next byte plus one in the receiver's file, without advancing the
 current pointer.  Returns nil if an error occurs."

| pos result |

pos := self position.
result := self next; next.
self position: pos.
^result
%

category: 'Positioning'
method: GsFile
position
  
"Returns the current position of the receiver's file, or nil if an error
 occurs."

^ self userAction: #GsfPosition onClient: isClient
%

category: 'Positioning'
method: GsFile
position: offset

"Changes the receiver's position in its file by the given offset, which may be
 negative or zero.  Returns the new position, or nil if an error occurs."

^ self _seekTo: offset opcode: 0 .
%

category: 'Writing'
method: GsFile
printBytes: byteObj

"Prints the bytes from the given byte object in decimal notation with
 line breaks to keep output lines from being too long.

 Returns the receiver, or nil if an error occurs.

 See also the method next:byteStringsInto:."

^ self userAction: #GsfPrintBytes onClient: isClient with: byteObj
%

category: 'Positioning'
method: GsFile
rewind

"Repositions the receiver's file to the beginning.  Returns 0, or nil if an
 error occurs."

self position: 0
%

category: 'Positioning'
method: GsFile
seekFromBeginning: offset

"Moves the receiver's position in its file to the given offset from the
 beginning of the file, which may be positive or zero, but not negative.
 Returns the new position, or nil if an error occurs."

^ self _seekTo: offset opcode: 0
%

category: 'Positioning'
method: GsFile
seekFromCurrent: offset

"Changes the receiver's position in its file by the given offset, which
 may be negative or zero.  Returns the new position, or nil if an error occurs."

^ self _seekTo: offset opcode: 1 
%

category: 'Positioning'
method: GsFile
seekFromEnd: offset

"Moves the receiver's position in its file to the given offset from the
 end of the file, which may be negative or zero, but not positive.
 Returns the new position, or nil if an error occurs."

^ self _seekTo: offset opcode: 2
%

category: 'Reading'
method: GsFile
skip: count

"Changes the receiver's position in its file by the given offset.  Offset may be
 positive, zero or negative.  Negative skipping past the beginning is not allowed.
 Returns the new position, or nil if an error occurs."

| pos |
pos := self position.
^ self position: pos + count
%

category: 'Private'
method: GsFile
_newStdFile: stdId isClient: clientBool

"Initialize the receiver to represent a standard file on the client."

clientBool ifTrue:[ isClient := true ] ifFalse:[ isClient := false ].
stdId == 0 ifTrue: [mode := 'r'. pathName := 'stdin']
          ifFalse:[mode := 'w'.
                   stdId == 1 ifTrue: [pathName := 'stdout']
                             ifFalse:[pathName := 'stderr'].
                  ].

%

category: 'Writing'
method: GsFile
write: byteObj itemCount: itemCount ofSize: bytesPerItem

"Writes itemCount items of size bytesPerItem from the given byte object to
 the receiver's file.  bytesPerItem must be between 1 and 4096 inclusive.
 Returns a count of bytes written, or nil if an error occurs."

^ self userAction: #GsfWrite onClient: isClient
       with: byteObj with: bytesPerItem with: itemCount
%

category: 'Private'
method: GsFile
_seekTo: offset opcode: seekKind

"seekKind  function
   0        seek to  start of file + offset
   1        seek  offset from current position
   2        seek to  end of file - offset

 Returns the new position, or nil if an error occurs."

^ self userAction: #GsfSeek onClient: isClient with: offset with: seekKind
%

category: 'Comparing'
method: GsFile
hash

"Returns a SmallInteger related to the value of the receiver.  If two instances
 of GsFile are equal (as compared by the = method), then they must have the same
 hash value."

^ self id hash
%

category: 'Comparing'
method: GsFile
= aFile

"Returns true if the receiver and aFile represent the same file system file.
 Returns false otherwise."

| idSelf |
(self == aFile) ifTrue: [^true].
aFile == nil ifTrue: [^false].
(aFile isKindOf: self class) ifFalse: [^false].
idSelf := self id.
(idSelf == 0) ifTrue: [^false].
^idSelf = (aFile id)
%

category: 'Accessing'
method: GsFile
id

"Returns the C-language file pointer to the standard I/O used by the receiver.
 Returns 0 if none exists."

 ^ self userAction: #GsfGetId onClient: isClient
%

category: 'Comparing'
method: GsFile
~= aFile

"Returns false if the receiver and aFile represent the same file system file.
 Returns true otherwise."

^ (self = aFile) not
%

category: 'Private'
method: GsFile
userAction: actionName onClient: onClient

"Executes a GsFile user action, passing it the following arguments:
    self
    any arguments to the primitive (see other uses of primitive 396)
 Maximum of 6 arguments (6 with: keywords) for this primitive."

<primitive: 396>

^ self _primitiveFailed: #userAction:onClient:
%

category: 'Private'
method: GsFile
userAction: actionName onClient: onClient with: arg1

"See GsFile | userAction:onClient: for documentation."

<primitive: 396>

^ self _primitiveFailed: #userAction:onClient:with:
%

category: 'Private'
method: GsFile
userAction: actionName onClient: onClient with: arg1 with: arg2

"See GsFile | userAction:onClient: for documentation."

<primitive: 396>

^ self _primitiveFailed: #userAction:onClient:with:with:
%

category: 'Private'
method: GsFile
userAction: actionName onClient: onClient with: arg1 with: arg2 with: arg3

"See GsFile | userAction:onClient: for documentation."

<primitive: 396>

^ self _primitiveFailed: #userAction:onClient:with:with:with:
%
